open List
open Cmdline
open Interactive
open Automation



let version () = print_string ("\
  LEO-II version v1.4.3\
  (compiled on " ^ Sys.os_type ^ " with Ocaml-" ^ Sys.ocaml_version ^ ")\n");

type arg =
  | FILENAME of string (*name of file to process*)
  (*the rest of the arguments are described in "help ()" below*)
  | ATP of string
  | ATPRC of string
  | ATPTIMEOUT of int
  | DEBUG of int
  | DIR of string
  | FOATP of string
  | HELP
  | INTERACTIVE
  | NOSLICES
  | MAXUNIDEPTH of int (*FIXME unidepth*)
  | PRETTYPRINTONLY (*FIXME prettyPrintOrig*)
  | PRIMSUBST of int
  | PROOFOUTPUT of int
  | RELEVANCEFILTER of int
  | SCRIPTMODE
  | SOS
  | TIMEOUT of int
  | TMPPATH of string (*FIXME tmp*)
  | UNFOLDDEFSEARLY
  | UNFOLDDEFSLATE
  | VERBOSE
  | VERSION
  | WRITEFOCLAUSES

(*Leo-II's configuration record*)
type configuration = {
  mutable foatp : string;
  mutable dir : string;
  mutable interactive : bool;
  mutable time_slices : int;
  mutable debug : int;
  mutable problemfiles : string list;
}
let global_conf : configuration = {
  foatp = "e";
  dir = "";
  interactive = false;
  time_slices = 4; 
  debug = 1;
  problemfiles = [];
}

let help () = print_string ("\
    Usage: leo [OPTIONS] [FILE]\n\
    Options:\n \
     --atp ATP=EXEC             Set the exec file for external prover ATP to EXEC\n \
                                (overrides the .leoatprc file, option can be used repeatedly)\n \
     --atprc FILE               Set ATP config file\n \
     --atptimeout N, -at N      Set the ATPtimeout (calls to E) to N seconds\n \
                                Default: 30s sec\n \
     --debug N, -D N            Set debug level to N\n \
                                (0 = no output, 1 = minimal output, 2 = full output)\n \
                                Default: 0\n \
     --dir DIR, -d DIR          Run on all files in DIR\n \
     --foatp PROVER, -f PROVER  Select PROVER as first-order prover\n \
                                Currently supported: E, Spass, Vampire, Gandalf\n \
                                Default prover is E (suggested)\n \
     --help, -h                 Display this help screen\n \
     --interactive, -i          Start interactive mode\n \
                                Default is non-interactive\n \
     --noslices -ns             Deactivates automatic time slicing; please use after the -t option e.g. as in leo -t 20 -ns file.p\n \ 
     --prettyPrintOrig, -ppo    Write pretty print of expanded original problem to a file (in tmp)\n \
     --primsubst N, -ps N       Set the prim subst level to N \n \
                                Default: 3\n \
     --proofoutput N, -po N     Print proof object \n \
                                N=0 (default): no proof output  \n \
                                N=1: Print proof object without detailed information on the fo-atp contribution \n \
                                N=2: Print proof object with detailed information on the fo-atp contribution \n \
     --relevancefilter N, -rf N Set the relevance filter to N\n \
                                Default: 0\n \
     --scriptmode, -s           Start script mode\n \
     --sos, -S                  (this flag is currently deactivated, don't use)\n \
     --timeout N, -t N          Enforce timeout after N seconds\n \
                                (0 = no timeout)\n \
                                Default: 600 sec\n \
     --tmp PATH, -tmp PATH      Set path for temporary files\n \
     --unfolddefsearly, -ude    Prefer early unfolding of definitions \n \
     --unfolddefslate, -udl     Prefer late unfolding of definitions \n \
     --unidepth N, -u N         Set the maximal (pre-)unification depth to N\n \
                                Default: 5\n \
     --verbose, -V              Run in verbose mode\n \
     --version, -v              Display version information\n \
     --writeFOclauses, -w       Write the first-order translations to a file (after termination)\n \
     \n")

let error msg =
  prerr_string ("LEO-II: " ^ msg ^ "\n");
  flush stderr;
  prerr_string ("Type \"" ^ Array.get Sys.argv 0 ^ " --help\" for argument info.\n");
  exit 255

(* try to read an integer from the command line arguments *)
let get_cl_int opt = function
  | [] -> error ("Option " ^ opt ^ " needs an argument.")
  | x :: xs ->
      try int_of_string x
      with _ -> error ("Option '" ^ opt ^ "' needs an argument.")

(* try to read a string from the command line arguments *)
let get_cl_string opt = function
  | [] -> error ("Option " ^ opt ^ " needs an argument.")
  | x :: xs -> x

(* parse command line arguments, return arg list. the arg list is accummulated
in "ps"*)
let rec parse_cl cs ps =
  match cs with
    | "--atp" :: xs ->
        parse_cl (tl xs) (ATP (get_cl_string (hd cs) xs) :: ps)
    | "--atprc" :: xs ->
        parse_cl (tl xs) (ATPRC (get_cl_string (hd cs) xs) :: ps)
    | "-at" :: xs
    | "--atptimeout" :: xs ->
        parse_cl (tl xs) (ATPTIMEOUT (get_cl_int (hd cs) xs) :: ps)
    | "-D" :: xs
    | "--debug" :: xs ->
        parse_cl (tl xs) (DEBUG (get_cl_int (hd cs) xs) :: ps)
    | "-d" :: xs
    | "--dir" :: xs ->
        parse_cl (tl xs) (DIR (get_cl_string (hd cs) xs) :: ps)
    | "-f" :: xs
    | "--foatp" :: xs ->
        parse_cl (tl xs) (FOATP (String.lowercase (get_cl_string (hd cs) xs)) :: ps)
    | "-h" :: xs
    | "--help" :: xs ->
        parse_cl xs (HELP :: ps)
    | "-i" :: xs
    | "--interactive" :: xs ->
        parse_cl xs (INTERACTIVE :: ps)
    | "-ns" :: xs
    | "--noslices" :: xs ->
        parse_cl xs (NOSLICES :: ps)
    | "-ppo" :: xs
    | "--prettyPrintOnly" :: xs ->
        parse_cl xs (PRETTYPRINTONLY :: ps)
    | "-ps" :: xs
    | "--primsubst" :: xs ->
        parse_cl  (tl xs) (PRIMSUBST (get_cl_int (hd cs) xs) :: ps)
    | "-po" :: xs
    | "--proofoutput" :: xs ->
        parse_cl (tl xs) (PROOFOUTPUT (get_cl_int (hd cs) xs) :: ps)
    | "-rf" :: xs
    | "--relevancefilter" :: xs ->
        parse_cl (tl xs) (RELEVANCEFILTER (get_cl_int (hd cs) xs) :: ps)
    | "-s" :: xs
    | "--scriptmode" :: xs ->
        parse_cl xs (SCRIPTMODE :: ps)
    | "-S" :: xs
    | "--sos" :: xs ->
        parse_cl xs (SOS :: ps)
    | "-t" :: xs
    | "--timeout" :: xs ->
        parse_cl (tl xs) (TIMEOUT (get_cl_int (hd cs) xs) :: ps)
    | "-tmp" :: xs
    | "--tmp" :: xs ->
        parse_cl (tl xs) (TMPPATH (get_cl_string (hd cs) xs) :: ps)
    | "-ude" :: xs
    | "--unfolddefsearly" :: xs ->
        parse_cl xs (UNFOLDDEFSEARLY :: ps)
    | "-udl" :: xs
    | "--unfolddefslate" :: xs ->
        parse_cl xs (UNFOLDDEFSLATE :: ps)
    | "-u" :: xs
    | "--unidepth" :: xs ->
        parse_cl (tl xs) (MAXUNIDEPTH (get_cl_int (hd cs) xs) :: ps)
    | "-V" :: xs
    | "--verbose" :: xs ->
        parse_cl xs (VERBOSE :: ps)
    | "-v" :: xs
    | "--version" :: xs ->
        parse_cl xs (VERSION :: ps)
    | "-w" :: xs
    | "--writeFOclauses" :: xs ->
        parse_cl xs (WRITEFOCLAUSES :: ps)
    | x :: xs ->
          if String.get x 0 = '-'
          then error ("Unknown command line argument \'" ^ x ^ "\'.")
          else
            (*the argument must be a filename, so attempt to parse it.*)
            parse_cl xs (FILENAME x :: ps)
    | [] -> ps (*return the arglist*)

let execute_conf () =
  if global_conf.interactive
  then comint ()
  else
    ignore(Interactive.initialize ());
  if global_conf.dir <> "" then
    let
        cmd = "prove-directory-with-fo-atp " ^ global_conf.dir ^ " " ^
      global_conf.foatp
    in
      print_endline ("\nLEO-II: " ^ cmd);
      ignore (Cmdline.execute_command cmd)
  else
    if global_conf.problemfiles <> [] then
      let solved_flag = ref false in
      let tackle filename =
        let cmd = "read-problem-file " ^ filename in
          Util.sysout 1 ("\nLEO-II: " ^ cmd);
          ignore (Cmdline.execute_command cmd);
	  let cmd = "call-fo-atp-early " ^ global_conf.foatp in
            Util.sysout 1 ("\nLEO-II: " ^ cmd);
            solved_flag := Cmdline.execute_command cmd;
	    if !solved_flag then ()
	    else 
              ( 
	      let no_of_axioms = (analyze_problem !leo_state) in
		Util.sysout 2 ("\n No.of.Axioms: " ^ string_of_int no_of_axioms ^ "\n");
		
		if global_conf.time_slices = 1 then
		  (let cmd = "read-problem-file " ^ filename in
		     Util.sysout 1 ("\nLEO-II: " ^ cmd);
		     ignore (Cmdline.execute_command cmd);
		     let cmd = "prove-with-fo-atp " ^ global_conf.foatp in
		       Util.sysout 1 ("\nLEO-II: " ^ cmd);
		       solved_flag := Cmdline.execute_command cmd)
		else
		  (if no_of_axioms <= 1 then
		     (if !solved_flag then ()
		      else
			((* no.1a *)
			  let cmd = "read-problem-file " ^ filename in
			    Util.sysout 1 ("\nLEO-II: " ^ cmd);
			    ignore (Cmdline.execute_command cmd);
			    let cmd = "prove-with-fo-atp " ^ global_conf.foatp in
			      Util.sysout 1 ("\nLEO-II: " ^ cmd);
			      solved_flag := Cmdline.execute_command cmd);
		      if !solved_flag then ()
		      else
			((* no.5 *)
			  let cmd = "flag-relevance-filter 0" in
			    Util.sysout 1 ("\nLEO-II: " ^ cmd);
			    ignore (Cmdline.execute_command cmd);
			    let cmd = "flag-prim-subst 0" in
			      Util.sysout 1 ("\nLEO-II: " ^ cmd);
			      ignore (Cmdline.execute_command cmd);
			      let cmd = "flag-max-uni-depth 8" in
				Util.sysout 1 ("\nLEO-II: " ^ cmd);
				ignore (Cmdline.execute_command cmd);
				let cmd = "flag-unfold-defs-early" in (* on *)
				  Util.sysout 1 ("\nLEO-II: " ^ cmd);
				  ignore (Cmdline.execute_command cmd);
				  let cmd = "read-problem-file " ^ filename in
				    Util.sysout 1 ("\nLEO-II: " ^ cmd);
				    ignore (Cmdline.execute_command cmd);
				    let cmd = "prove-with-fo-atp " ^ global_conf.foatp in
				      Util.sysout 1 ("\nLEO-II: " ^ cmd);
				      solved_flag := Cmdline.execute_command cmd);
		      if !solved_flag then ()
		      else
			((* no.5 *)
			  let cmd = "flag-relevance-filter 0" in
			    Util.sysout 1 ("\nLEO-II: " ^ cmd);
			    ignore (Cmdline.execute_command cmd);
			    let cmd = "flag-prim-subst 1" in
			      Util.sysout 1 ("\nLEO-II: " ^ cmd);
			      ignore (Cmdline.execute_command cmd);
			      let cmd = "flag-max-uni-depth 1" in
				Util.sysout 1 ("\nLEO-II: " ^ cmd);
				ignore (Cmdline.execute_command cmd);
				let cmd = "read-problem-file " ^ filename in
				  Util.sysout 1 ("\nLEO-II: " ^ cmd);
				  ignore (Cmdline.execute_command cmd);
				  let cmd = "prove-with-fo-atp " ^ global_conf.foatp in
				    Util.sysout 1 ("\nLEO-II: " ^ cmd);
				    solved_flag := Cmdline.execute_command cmd);
		      if !solved_flag then ()
		      else
			((* no.5 *)
			  let cmd = "flag-relevance-filter 0" in
			    Util.sysout 1 ("\nLEO-II: " ^ cmd);
			    ignore (Cmdline.execute_command cmd);
			    let cmd = "flag-prim-subst 3" in
			      Util.sysout 1 ("\nLEO-II: " ^ cmd);
			      ignore (Cmdline.execute_command cmd);
			      let cmd = "flag-max-uni-depth 3" in
				Util.sysout 1 ("\nLEO-II: " ^ cmd);
				ignore (Cmdline.execute_command cmd);
				let cmd = "read-problem-file " ^ filename in
				  Util.sysout 1 ("\nLEO-II: " ^ cmd);
				  ignore (Cmdline.execute_command cmd);
				  let cmd = "prove-with-fo-atp " ^ global_conf.foatp in
				    Util.sysout 1 ("\nLEO-II: " ^ cmd);
				    solved_flag := Cmdline.execute_command cmd);
		     )
		   else
		     if no_of_axioms > 100 then
		       (if !solved_flag then ()
			else
			  ((* no.3 *)
			    let cmd = "flag-relevance-filter 2" in
			    let _ = Util.sysout 1 ("\nLEO-II: " ^ cmd) in
			      ignore (Cmdline.execute_command cmd);
			      let cmd = "flag-prim-subst 2" in
				ignore (Cmdline.execute_command cmd);
				let cmd = "flag-max-uni-depth 5" in
				  Util.sysout 1 ("\nLEO-II: " ^ cmd);
				  ignore (Cmdline.execute_command cmd);
				  let cmd = "read-problem-file " ^ filename in
				    Util.sysout 1 ("\nLEO-II: " ^ cmd);
				    ignore (Cmdline.execute_command cmd);
				    let cmd = "prove-with-fo-atp " ^ global_conf.foatp in
				      Util.sysout 1 ("\nLEO-II: " ^ cmd);
				      solved_flag := Cmdline.execute_command cmd);
			if !solved_flag then ()
			else
			  ((* no.3 *)
			    let cmd = "flag-relevance-filter -1" in
			      Util.sysout 1 ("\nLEO-II: " ^ cmd);
			      ignore (Cmdline.execute_command cmd);
			      let cmd = "flag-prim-subst 0" in
				ignore (Cmdline.execute_command cmd);
				let cmd = "flag-max-uni-depth 1" in
				  Util.sysout 1 ("\nLEO-II: " ^ cmd);
				  ignore (Cmdline.execute_command cmd);
				  let cmd = "read-problem-file " ^ filename in
				    Util.sysout 1 ("\nLEO-II: " ^ cmd);
				    ignore (Cmdline.execute_command cmd);
				    let cmd = "prove-with-fo-atp " ^ global_conf.foatp in
				      Util.sysout 1 ("\nLEO-II: " ^ cmd);
				      solved_flag := Cmdline.execute_command cmd);
			if !solved_flag then ()
			else
			  ((* no.4 *)
			    let cmd = "flag-relevance-filter 0" in
			      Util.sysout 1 ("\nLEO-II: " ^ cmd);
			      ignore (Cmdline.execute_command cmd);
			      let cmd = "flag-prim-subst 3" in
				Util.sysout 1 ("\nLEO-II: " ^ cmd);
				ignore (Cmdline.execute_command cmd);
				let cmd = "flag-max-uni-depth 3" in
				  Util.sysout 1 ("\nLEO-II: " ^ cmd);
				  ignore (Cmdline.execute_command cmd);
				  let cmd = "read-problem-file " ^ filename in
				    Util.sysout 1 ("\nLEO-II: " ^ cmd);
				    ignore (Cmdline.execute_command cmd);
				    let cmd = "prove-with-fo-atp " ^ global_conf.foatp in
				      Util.sysout 1 ("\nLEO-II: " ^ cmd);
				      solved_flag := Cmdline.execute_command cmd);
			if !solved_flag then ()
			else
			  ((* no.3 *)
			    let cmd = "flag-relevance-filter 1" in
			      Util.sysout 1 ("\nLEO-II: " ^ cmd);
			      ignore (Cmdline.execute_command cmd);
			      let cmd = "flag-prim-subst 0" in
				ignore (Cmdline.execute_command cmd);
				let cmd = "flag-max-uni-depth 2" in
				  Util.sysout 1 ("\nLEO-II: " ^ cmd);
				  ignore (Cmdline.execute_command cmd);
				  let cmd = "flag-unfold-defs-early" in (* on *)
				    Util.sysout 1 ("\nLEO-II: " ^ cmd);
				    ignore (Cmdline.execute_command cmd);
				    let cmd = "read-problem-file " ^ filename in
				      Util.sysout 1 ("\nLEO-II: " ^ cmd);
				      ignore (Cmdline.execute_command cmd);
				      let cmd = "prove-with-fo-atp " ^ global_conf.foatp in
					Util.sysout 1 ("\nLEO-II: " ^ cmd);
					solved_flag := Cmdline.execute_command cmd);
		       )
		     else
		       (if !solved_flag then ()
			else
			  ((* no.1 *)
			    let cmd = "read-problem-file " ^ filename in
			      Util.sysout 1 ("\nLEO-II: " ^ cmd);
			      ignore (Cmdline.execute_command cmd);
			      let cmd = "prove-with-fo-atp " ^ global_conf.foatp in
				Util.sysout 1 ("\nLEO-II: " ^ cmd);
				solved_flag := Cmdline.execute_command cmd);
			if !solved_flag then ()
			else
			  ((* no.5 *)
			    let cmd = "flag-prim-subst 1" in
			      Util.sysout 1 ("\nLEO-II: " ^ cmd);
			      ignore (Cmdline.execute_command cmd);
			      let cmd = "flag-max-uni-depth 1" in
				Util.sysout 1 ("\nLEO-II: " ^ cmd);
				ignore (Cmdline.execute_command cmd);
				let cmd = "read-problem-file " ^ filename in
				  Util.sysout 1 ("\nLEO-II: " ^ cmd);
				  ignore (Cmdline.execute_command cmd);
				  let cmd = "prove-with-fo-atp " ^ global_conf.foatp in
				    Util.sysout 1 ("\nLEO-II: " ^ cmd);
				    solved_flag := Cmdline.execute_command cmd);
			if !solved_flag then ()
			else
			  ((* no.3 *)
			    let cmd = "flag-relevance-filter 2" in
			      Util.sysout 1 ("\nLEO-II: " ^ cmd);
			      ignore (Cmdline.execute_command cmd);
			      let cmd = "flag-prim-subst 0" in
				ignore (Cmdline.execute_command cmd);
				let cmd = "flag-max-uni-depth 1" in
				  Util.sysout 1 ("\nLEO-II: " ^ cmd);
				  ignore (Cmdline.execute_command cmd);
				  let cmd = "read-problem-file " ^ filename in
				    Util.sysout 1 ("\nLEO-II: " ^ cmd);
				    ignore (Cmdline.execute_command cmd);
				    let cmd = "prove-with-fo-atp " ^ global_conf.foatp in
				      Util.sysout 1 ("\nLEO-II: " ^ cmd);
				      solved_flag := Cmdline.execute_command cmd);
			if !solved_flag then ()
			else
			  ((* no.3 *)
			    let cmd = "flag-relevance-filter -1" in
			      Util.sysout 1 ("\nLEO-II: " ^ cmd);
			      ignore (Cmdline.execute_command cmd);
			      let cmd = "flag-prim-subst 0" in
				ignore (Cmdline.execute_command cmd);
				let cmd = "flag-max-uni-depth 4" in
				  Util.sysout 1 ("\nLEO-II: " ^ cmd);
				  ignore (Cmdline.execute_command cmd);
				  let cmd = "read-problem-file " ^ filename in
				    Util.sysout 1 ("\nLEO-II: " ^ cmd);
				    ignore (Cmdline.execute_command cmd);
				    let cmd = "prove-with-fo-atp " ^ global_conf.foatp in
				      Util.sysout 1 ("\nLEO-II: " ^ cmd);
				      solved_flag := Cmdline.execute_command cmd);
		       )
		  )
	      )
      in
	List.iter tackle global_conf.problemfiles
    else error ("No problem file given.")
      
(* process the argument list, build the configuration record
 * and finally call execute_conf *)
let rec process args = match args with
  | FILENAME s :: args ->
      global_conf.problemfiles <- s :: global_conf.problemfiles;
      process args
  | ATP s :: args ->
      let eqpos = String.index s '=' in
      let len = String.length s in
      let atp = String.sub s 0 eqpos in
      let executable = String.sub s (eqpos + 1) (len - eqpos - 1) in
      (* print_string ("ATP: "^atp^", exec: "^executable^"\n"); *)
      if Sys.file_exists executable 
      then
	( Automation.atp_cmds := (atp, executable) :: !Automation.atp_cmds;
	  Automation.atp_configured := true;
	  Util.sysout 1 ("  Configured: " ^ atp ^ " = " ^ executable ^ "\n");
	);
	process args
  | ATPRC s :: args ->
      Automation.atp_config_file := s;
      process args
  | ATPTIMEOUT n :: args ->
      ignore(Main.set_flag_atp_timeout Main.state_initialize (max 2 n));
      process args
  | DEBUG n :: args ->
      global_conf.debug <- (min 2 (max 0 n));
      Util.debuglevel := (min 2 (max 0 n));
      process args
  | DIR s :: args ->
      global_conf.dir <- s;
      process args
  | FOATP s :: args ->
      if not (s = "e" || s = "spass" || s = "vampire")
      then error ("Unknown foatp \"" ^ s ^ "\".")
      else
        global_conf.foatp <- s;
        process args
  | HELP :: args ->
      help ();
      process args
  | INTERACTIVE :: args ->
      global_conf.interactive <- true;
      Cmdline.interactive := true;
      Util.debuglevel := 1;
      Cmdline.save_tcio ();
      process args
  | NOSLICES :: args ->
      global_conf.time_slices <- 1;
      process args
  | MAXUNIDEPTH n :: args ->
      ignore(Main.set_flag_max_uni_depth Main.state_initialize (max 1 n));
      process args
  | PRETTYPRINTONLY :: args ->
      ignore(Main.set_flag_pretty_print_only Main.state_initialize true);
      process args
  | PRIMSUBST n :: args ->
      ignore(Main.set_flag_prim_subst Main.state_initialize n);
      process args
  | PROOFOUTPUT n :: args ->
      ignore(Main.set_flag_proof_output Main.state_initialize n);
      process args
  | RELEVANCEFILTER n :: args ->
      ignore(Main.set_flag_relevance_filter Main.state_initialize (max 0 n));
      process args
  | SCRIPTMODE :: args ->
      global_conf.interactive <- true;
      process args
  | SOS :: args ->
      ignore(Main.set_flag_sos Main.state_initialize false);
      process args
  | TIMEOUT n :: args ->
      let globaltmo  = (max 1 n) in
      let atptmo = (min 25 (max 1 (n / (global_conf.time_slices + 2)))) in 
      let localtmo = (max 1 (n / global_conf.time_slices)) in
       Interactive.set_original_timeout globaltmo;
       Interactive.set_timeout localtmo;
       ignore
        (Main.set_flag_atp_timeout
         Main.state_initialize atptmo);
       ignore
        (Main.set_flag_max_local_time
         Main.state_initialize localtmo);
       process args
  | TMPPATH s :: args ->
      Util.tmp_path := s;
      process args
  | UNFOLDDEFSEARLY :: args ->
      ignore(Main.set_flag_unfold_defs_early Main.state_initialize true);
      process args
  | UNFOLDDEFSLATE :: args ->
      ignore(Main.set_flag_unfold_defs_early Main.state_initialize false);
      process args
  | VERBOSE :: args ->
      ignore(Main.set_flag_verbose Main.state_initialize true);
      process args
  | VERSION :: args ->
      version ();
      process args
  | WRITEFOCLAUSES :: args ->
      ignore(Main.set_flag_write_fo_like_clauses Main.state_initialize true);
      process args
  | [] -> execute_conf ()

let cleanup_quit () =
  Interactive.kill_children ();
  Util.StringSet.iter Util.try_delete_file !Util.tmpfiles;
  exit 255

let main =
  (*register signal handlers*)
  let sigxcpu = 24 in (*POSIX signal number for "CPU time limit exceeded"*)
    ignore(Sys.signal Sys.sigquit (Sys.Signal_handle (fun _ -> raise Termination)));
    ignore(Sys.signal sigxcpu (Sys.Signal_handle (fun _ -> raise Termination)));
    ignore(Sys.signal Sys.sigchld Sys.Signal_ignore);
  try
    let com_line = List.tl (Array.to_list Sys.argv) in
    let args = parse_cl com_line [] in
    process args;
  with
    | Failure x ->
       prerr_string ("\nUnexpected error occurred:\n" ^ x ^ "\n");
    | Termination ->
       prerr_string ("\nTerminating\n");
    | e ->
       prerr_string ("\nError occurred:" ^ Printexc.to_string e ^ "\n");
  (*cleanup_quit ()*)
